@@ -6,6 +6,7 @@ import uuid
 from typing import Dict, List, Any, Tuple, Union
 from os.path import join
 from datetime import datetime
+from copy import deepcopy
 import requests
 
 from ia.gaius.genome_info import Genome
@@ -47,7 +48,7 @@ def _remove_unique_id(response: dict) -> dict:
 class AgentClient:
     """Interface for interacting with bottles."""
 
-    def __init__(self, bottle_info):
+    def __init__(self, bottle_info, verify=True):
         """
         Provide bottle information in a dictionary.
 
@@ -81,6 +82,7 @@ class AgentClient:
         self.gaius_agent = None
         self.send_unique_ids = True
         self.summarize_for_single_node = True
+        self._verify = verify
         if 'secure' not in self._bottle_info or self._bottle_info['secure']:
             self._secure = True
             if not self.name:
@@ -115,7 +117,7 @@ class AgentClient:
 
     def connect(self) -> Dict:
         """Establishes initial connection to GAIuS agent and grabs the bottle's gaius_agent's genome for node definitions."""
-        response_data = self.session.get(self._url + 'connect', verify=self._secure, headers=self._headers).json()
+        response_data = self.session.get(self._url + 'connect', verify=self._verify, headers=self._headers).json()
         if 'status' not in response_data or response_data['status'] != 'okay':
             self._connected = False
             raise AgentConnectionError("Connection failed!", response_data)
@@ -167,9 +169,9 @@ class AgentClient:
             full_path = f'{self._url}{node["id"]}/{path}'
             try:
                 if data is not None:
-                    response = query_method(full_path, verify=self._secure, headers=self._headers, json={'data': data})
+                    response = query_method(full_path, verify=self._verify, headers=self._headers, json={'data': data})
                 else:
-                    response = query_method(full_path, verify=self._secure, headers=self._headers)
+                    response = query_method(full_path, verify=self._verify, headers=self._headers)
                 response.raise_for_status()
                 response = response.json()
                 if response['status'] != 'okay':
@@ -287,7 +289,7 @@ class AgentClient:
         for node in nodes:
             response = self.session.post(
                 f"{self._url}{node['id']}/genes/change",
-                verify=self._secure,
+                verify=self._verify,
                 headers=self._headers,
                 json={'data': gene_data},
             ).json()
@@ -311,7 +313,7 @@ class AgentClient:
         for node in nodes:
             try:
                 response = self.session.get(
-                    f"{self._url}{node['id']}/gene/{gene}", verify=self._secure, headers=self._headers
+                    f"{self._url}{node['id']}/gene/{gene}", verify=self._verify, headers=self._headers
                 ).json()
                 if 'error' in response or response['status'] == 'failed':
                     if len(nodes) == 1 and self.summarize_for_single_node:
@@ -339,7 +341,7 @@ class AgentClient:
         for node in nodes:
             try:
                 response = self.session.get(
-                    f"{self._url}{node['id']}/model/{model_name}", headers=self._headers, verify=self._secure
+                    f"{self._url}{node['id']}/model/{model_name}", headers=self._headers, verify=self._verify
                 ).json()
                 if 'error' in response or response['status'] == 'failed':
                     if len(nodes) == 1 and self.summarize_for_single_node:
@@ -368,7 +370,7 @@ class AgentClient:
         for node in nodes:
             try:
                 response = self.session.delete(
-                    f"{self._url}{node['id']}/model/{model_name}", headers=self._headers, verify=self._secure
+                    f"{self._url}{node['id']}/model/{model_name}", headers=self._headers, verify=self._verify
                 ).json()
                 if 'error' in response or response['status'] == 'failed':
                     if len(nodes) == 1 and self.summarize_for_single_node:
@@ -397,7 +399,7 @@ class AgentClient:
         for node in nodes:
             try:
                 response = self.session.put(
-                    f"{self._url}{node['id']}/model/{model_name}", headers=self._headers, verify=self._secure,
+                    f"{self._url}{node['id']}/model/{model_name}", headers=self._headers, verify=self._verify,
                     json={'model': model}
                 ).json()
                 if 'error' in response or response['status'] == 'failed':
@@ -427,7 +429,7 @@ class AgentClient:
         for node in nodes:
             try:
                 response = self.session.get(
-                    f"{self._url}{node['id']}/model/{model_name}", headers=self._headers, verify=self._secure
+                    f"{self._url}{node['id']}/model/{model_name}", headers=self._headers, verify=self._verify
                 ).json()
                 if 'error' in response or response['status'] == 'failed':
                     if len(nodes) == 1 and self.summarize_for_single_node:
@@ -464,7 +466,7 @@ class AgentClient:
         for node in nodes:
             try:
                 response = self.session.get(
-                    f"{self._url}{node['id']}/vector", headers=self._headers, verify=self._secure, json={'data': vector_name}
+                    f"{self._url}{node['id']}/vector", headers=self._headers, verify=self._verify, json={'data': vector_name}
                 ).json()
                 if 'error' in response or response['status'] == 'failed':
                     if len(nodes) == 1 and self.summarize_for_single_node:
@@ -490,7 +492,7 @@ class AgentClient:
             try:
                 response = self.session.post(
                     f"{self._url}{node['id']}/gene/increment-recall-threshold",
-                    verify=self._secure,
+                    verify=self._verify,
                     headers=self._headers,
                     json={'increment': increment},
                 ).json()
@@ -543,7 +545,7 @@ class AgentClient:
             nodes = [node for node in self.all_nodes if (node['name'] in nodes)]
             results = {}
             for node in nodes:
-                response = self.session.get(f"{self._url}{node['id']}/ping", verify=self._secure, headers=self._headers).json()
+                response = self.session.get(f"{self._url}{node['id']}/ping", verify=self._verify, headers=self._headers).json()
                 if 'error' in response or response["status"] == 'failed':
                     if len(nodes) == 1 and self.summarize_for_single_node:
                         raise Exception("Request Failure:", response)
@@ -564,7 +566,7 @@ class AgentClient:
                     f"{self._url}database", 
                     headers=_headers,
                     json={},
-                    verify=self._secure
+                    verify=self._verify
                 )
         archive_file = join(directory, f'{self.gaius_agent}-{self.name}-{datetime.now()}-kb.archive.gz')
         with open(archive_file, 'wb') as f:
@@ -572,7 +574,7 @@ class AgentClient:
         return f"Saved as {archive_file}"
     
     @_ensure_connected
-    def get_kbs_as_json(self, nodes: List = None, directory='./', separated=False, ids=True):
+    def get_kbs_as_json(self, nodes: List = None, directory='./', filename=None, separated=False, ids=True, obj=False):
         """Returns KBs of specified nodes as JSON Objects
 
         Args:
@@ -591,7 +593,13 @@ class AgentClient:
         if isinstance(output, dict):
             # print(f'nodes = {nodes}, output = {output.keys()}')
             altered_output = {}
-            if ids == True:
+            if len(self.all_nodes) == 1:
+                print(f'topology only has one node {self.all_nodes[0]["id"]}')
+                if ids == True:
+                    altered_output = {self.all_nodes[0]["id"] : output}
+                else:
+                    altered_output = {"P1" : output}
+            elif ids == True and "metadata" not in output:
                 for key in output.keys():
                     node_id = [node["id"] for node in self.all_nodes if node['name'] == key][0]
                     altered_output[node_id] = output[key]
@@ -605,18 +613,26 @@ class AgentClient:
                         f.write(json.dumps({key: altered_output[key]}))
             else:
                 archive_file = join(directory, f'{self.gaius_agent}-{self.name}-{datetime.now()}-kb.json')
+                if filename:
+                    archive_file = filename
+                    
                 filenames.append(archive_file)
+                
+                # return kb as json object
+                if obj == True:
+                    return altered_output
+                # otherwise write json file
                 with open(archive_file, 'w') as f:
                     f.write(json.dumps(altered_output))
                 
         return f"saved kbs as {', '.join(filenames)}"
     
     @_ensure_connected
-    def load_kbs_from_json(self, path):
+    def load_kbs_from_json(self, path=None, obj=None):
         """Load KBs from a JSON file
 
         Args:
-            path (str, required): _description_. path to JSON file where KB is stored.
+            path (str, required): _description_. path to JSON file where KBs are stored.
             If primitive names (P1, P2, P3, etc.)
 
         Returns:
@@ -629,27 +645,34 @@ class AgentClient:
             node_dict[item['name']] = item['id']
             node_dict[item['id']] = item['id']
         
-        with open(path, 'r') as f:
-            kb = json.load(f)
-            for key, value in kb.items():
-                prim_id = key
-                if prim_id in node_dict:
-                    prim_id = node_dict[prim_id]
-                else:
-                    print(f'Warning, node {prim_id} not found in topology, skipping')
-                    continue
-                print(f'loading node {prim_id} from kb for {key}')
-                # print(f'value = {value}')
-                print(f'Url = {self._url}{prim_id}/load_kb')
-                response = self.session.post(f'{self._url}{prim_id}/load_kb', headers={'X-API-KEY' : 'ABCD-1234'}, json={'data': value})
-                # print(response)
-                if response.status_code != 200:
-                    print(f'error loading kb for node {prim_id}: {response.text}')
-                    return 'failed', response.status_code
-                print(f'loading node {prim_id} succeeded')
-        # except Exception as e:
-        #     print(e.args)
-        #     return 'failed'
+        if path is not None:
+            with open(path, 'r') as f:
+                kb = json.load(f)
+        elif obj is not None:
+            kb = obj
+        else:
+            raise Exception("Must specify path or obj argument")
+        
+        # check if the kb is only for a single node (simple-topology edge case)
+        if 'metadata' in kb.keys():
+            kb = {'P1' : kb}
+
+        for key, value in kb.items():
+            prim_id = key
+            if prim_id in node_dict:
+                prim_id = node_dict[prim_id]
+            else:
+                print(f'Warning, node {prim_id} not found in topology, skipping')
+                continue
+            print(f'loading node {prim_id} from kb for {key}')
+            # print(f'value = {value}')
+            # print(f'Url = {self._url}{prim_id}/load_kb')
+            response = self.session.post(f'{self._url}{prim_id}/load_kb', headers={'X-API-KEY' : self._api_key}, json={'data': value})
+            # print(response)
+            if response.status_code != 200:
+                print(f'error loading kb for node {prim_id}: {response.text}')
+                return 'failed', response.status_code
+            print(f'loading node {prim_id} succeeded')
         
         return 'success'
     
@@ -664,7 +687,7 @@ class AgentClient:
             data = f.read()
             response = self.session.put(f'{self._url}database', 
                         headers=_headers,
-                        verify=self._secure,
+                        verify=self._verify,
                         data=data)
         return response.json()
 
@@ -688,5 +711,101 @@ class AgentClient:
     def tf_attach(self, name, config):
         return self.session.post(f'{self._url}/thinkflux/{name}/attach', data={'config' : config})
         
+    @_ensure_connected
+    def investigate_record(self, node, record):
+        data = {}
+        try: 
+            if isinstance(node, list):
+                node = node[0]
+            try:
+                # if the node passed is a primitive_id, get the name (P1, P2, etc.) of the node
+                node = self.genome.primitives[node]['name']
+            except:
+                pass
+            # print(f"investigating record {record}")
+            return self.__get_model_details(model=record, node=[node], topLevel=True)
+
+        except Exception as e:
+            print(str(e))
+        return data
+
+    @_ensure_connected
+    def __get_model_details(self, model, node, topLevel=False):
+        """Recursive function to get details about a specific model on a specific node
+
+        Args:
+            model (str): the model hash to look for
+            node (list(str)): The node to look for the model on (provided as a list with a single element)
+            topLevel (bool, optional): flag used to show that the function is at the top level (no recursing). Defaults to False.
+
+        Returns:
+            dict: recursive structure representing the data from the topLevel model
+        """
+        if model == None:
+            return None
+        p_id = self.genome.get_primitive_map()[node[0]]
+        if topLevel is True:
+            # top level request
+            record_data = {'record' : f"{model}",
+                        'subitems':[],
+                        'model': self.get_model(model, nodes=node),
+                        'node': p_id,
+                        'topLevel' : topLevel,
+                        'bottomLevel' : False,
+            }
+        else:            
+            record_data = {'record' : f"PRIMITIVE|{p_id}|{model['name']}",
+                        'subitems':[],
+                        'model': model,
+                        'node': p_id,
+                        'topLevel' : topLevel,
+                        'bottomLevel' : False
+            }
+        for event_list in record_data['model']['sequence']:
+            event = []
+            for item in event_list:
+                symbol = {}
+                split_model = item.split('|')
+                if split_model[0] == 'PRIMITIVE':
+                    node_id = split_model[1]
+                    try:
+                        node_name = self.genome.primitives[node_id]['name']
+                    except:
+                        # node name already provided
+                        node_name = node_id
+                        pass
+                    sub_model = self.get_model(split_model[2], nodes=[node_name])
+                    symbol = self.__get_model_details(model=sub_model, node=[node_name])
+                elif split_model[0] == 'VECTOR':
+                    symbol = {'record': item,
+                                'data'  : self.get_vector(item, nodes=node),
+                                'subitems': None,
+                                'bottomLevel': True,
+                                'topLevel' : False,
+                                'node':node[0]}
+                else:
+                    # print(f'reached symbol: {item}')
+                    if 'VECTOR' in item:
+                        symbol = {'record': item,
+                                    'data'  : self.get_vector(item, nodes=node),
+                                    'subitems': None,
+                                    'bottomLevel': True,
+                                    'topLevel' : False,
+                                    'node' : node[0]}
+                    else:
+                        symbol = item
+                event.append(deepcopy(symbol))
+            record_data["subitems"].append(event)
+        if tuple(record_data["subitems"]) == tuple(record_data["model"]["sequence"]):
+            del record_data["subitems"]
+        else:
+            pass
+        if not topLevel:
+            if 'subitems' in record_data:
+                record_data["subitems"] = tuple(record_data["subitems"])
+            else:
+                record_data["subitems"] = None
+                record_data['bottomLevel'] = True
+        return record_data
     
 __all__ = (AgentConnectionError, AgentClient, AgentQueryError)
